import hashlib
import lib.serpent
from lib.serpent import Serpent
import config
import binascii
import time
import struct
import sys

PACKET_SIZE = 580 
MAX_MESSAGE_SIZE = 512
MAX_SECRET_SIZE = 24
TS_ACCEPTABLE_SKEW = 60 * 15

class Infosec(object): 
    def __init__(self, server=None):
        self.server = server

    def pack(self, peer, message, timestamp=None):
        # if we are rebroadcasting we need to use the original timestamp
        if(timestamp == None):
            int_ts = int(time.time())
        else:
            int_ts = timestamp

        serpent = Serpent(self._pad(peer.local_secret, MAX_SECRET_SIZE))
        padded_message = self._pad(message, MAX_MESSAGE_SIZE)
        ciphertext_bytes = serpent.encrypt(padded_message.encode("ascii"))
        digest = hashlib.sha512(peer.local_secret + ciphertext_bytes)
        digest_bytes = digest.digest()

        # we want to ignore this ts if it is sent back to us
        self.server.recent.insert(int_ts)
        self.server.print_debug("added %d to recent" % int_ts)

        # build the packet and return it
        packet = struct.pack("!L64s512s", int_ts, digest_bytes, ciphertext_bytes)
        return packet 
    
    def unpack(self, peer, packet):
        try:
            int_ts, digest, ciphertext_bytes = struct.unpack("!L64s512s", packet)
        except: 
            self.server.print_error("Discarding malformed packet?")
            return None, None

        # Check the timestamp and digest
        digest_check = hashlib.sha512(peer.remote_secret + ciphertext_bytes).digest()
        
        if(int_ts not in self._ts_range()):
            self.server.print_debug("rejected message with timestamp out of range")
            return None, None
        elif(self.server.recent.has(int_ts)):
            self.server.print_debug("rejected known message: %d" % int_ts)
            return None, None
        elif(digest_check != digest):
            self.server.print_debug("name: %s" % peer.name)
            self.server.print_debug("remote_secret: %s" % peer.remote_secret)
            self.server.print_debug("ciphertext_bytes: %s" % binascii.hexlify(ciphertext_bytes))
            self.server.print_debug("digest_check: %s" % binascii.hexlify(digest_check))
            self.server.print_debug("rejected bad digest: %s" % binascii.hexlify(digest))
            return None, None
        else:
            # Return the cleartext
            serpent = Serpent(self._pad(peer.remote_secret, MAX_SECRET_SIZE))
            cleartext = serpent.decrypt(ciphertext_bytes).rstrip()
            self.server.print_debug("received, validated, and decrypted udp packet: %s" % cleartext)
            self.server.recent.insert(int_ts)
            return cleartext, int_ts

    def _pad(self, text, size):
        return text.ljust(size)

    def _ts_range(self):
        current_ts = int(time.time())
        return range(current_ts - TS_ACCEPTABLE_SKEW, current_ts + TS_ACCEPTABLE_SKEW)
